/*
Roger's SE regex system adapted for use in Nigel's z and z32 pagers.

nib 1998-07-27
*/

#include <stdio.h>
#include <ctype.h>

void compile(char **cp);
int execute(int gf, char *buf, char* braslist[], char *braelist[],
						int ignore_case);
int advance(char *lp, char *ep, char *braslist[], char *braelist[],
						int ignore_case);
int cclass(char *set, int c, int af, int ignore_case);

#define EXSIZE	128
#define NBRA	5

char *BADPATTERN = "Illformed search pattern";

char expbuf [EXSIZE + 4];
int circfl;


#define	CCHR	1
#define	CDOT	2
#define	CCL	3
#define	NCCL	4

#define	STAR	4
#define PLUS	8

#define	CDOL	13
#define	CEOF	14
#define	CBRA	15
#define	CKET	16

#define RANGE	((char)0x7f)

void errfun(char *mess)
{
	putchar('G'-0x40);
	expbuf[0] = CEOF;
}

void compile(char **cp)
{
	int c;
	char *ep;
	char *lastep;
	char bracket[NBRA], *bracketp;
	int nbra;
	int cclcnt;

	ep = expbuf;
	bracketp = bracket;
	nbra = 1;
	if ((c = *(*cp)++) == '\0')
	{
		if (*ep == 0)
			errfun (BADPATTERN);
		if (c == '\0')
			--*cp;
		return;
	}
	circfl = 0;
	if (c == '^')
	{
		circfl++;
		c = *(*cp)++;
	}
	else if (c == '*' || c == '+')
		goto cerror;
	for (;;)
	{
		if (ep >= &expbuf [EXSIZE])
			goto cerror;
		if (c == '\0')
		{
			*ep++ = CEOF;
			if (c == '\0')
				--*cp;
			return;
		}
		if (c != '*' && c != '+')
			lastep = ep;
		switch (c)
		{
#if 0	/* Have no use for brackets at the moment - so be like vi and
	   don't have them as specials - nib.
	*/
		case '(':
			if (nbra >= NBRA)
				goto cerror;
			*bracketp++ = nbra;
			*ep++ = CBRA;
			*ep++ = nbra++;
			break;

		case ')':
			if (bracketp <= bracket)
				goto cerror;
			*ep++ = CKET;
			*ep++ = *--bracketp;
			break;
#endif

		case '.':
			*ep++ = CDOT;
			break;

		case '*':	/* zero or more occurrences */
			if (*lastep == CBRA || *lastep == CKET)
				errfun (BADPATTERN);
			*lastep += STAR;
			break;

		case '+':	/* one or more occurrences */
			if (*lastep == CBRA || *lastep == CKET)
				errfun (BADPATTERN);
			*lastep += PLUS;
			break;

		case '$':
			if (**cp != '\0')
				goto cerror;
			*ep++ = CDOL;
			break;

		case '[':
			*ep++ = CCL;
			*ep++ = 0;
			cclcnt = 1;
			if ((c = *(*cp)++) == '^')
			{
				c = *(*cp)++;
				ep[-2] = NCCL;
			}
			do
			{
				if (c == '\0' || c == RANGE)
					goto cerror;
				if (c == '-' && cclcnt > 1)	/* char range */
				{
					c = *--ep;	/* last char is start of range */
					*ep++ = RANGE;
					*ep++ = c;
					*ep++ = *(*cp)++;
					cclcnt += 2;	/* one's already counted */
				}
				else	/* single char */
				{
					*ep++ = c;
					cclcnt++;
				}
				if (ep >= &expbuf [EXSIZE])
					goto cerror;
			}
			while ((c = *(*cp)++) != ']');
			lastep [1] = cclcnt;
			break;

		case '\\':
			if ((c = *(*cp)++) == 0)
				goto cerror;
			*ep++ = CCHR;
			*ep++ = c;
			break;

		default:
			*ep++ = CCHR;
			*ep++ = c;
			break;
		}
		c = *(*cp)++;
	}
cerror:
	expbuf[0] = 0;
	errfun (BADPATTERN);
}



int execute(int gf, char *buf, char *braslist[], char *braelist[],
					int ignore_case)
{
	char *p1, *p2, c;

	if (gf)
	{
		if (circfl)
			return (0);
		p1 = braelist [0];
	}
	else
	{
		p1 = buf;
	}
	p2 = expbuf;
	if (circfl)
	{
		braslist [0] = p1;
		return (advance (p1, p2, braslist, braelist, ignore_case));
	}

	/* fast check for first character */

	if (*p2 == CCHR)
	{
		c = p2 [1];
		do
		{
			if (!ignore_case && *p1 != c || tolower(*p1) !=
						tolower(c))
				continue;
			if (advance (p1, p2, braslist, braelist, ignore_case))
			{
				braslist [0] = p1;
				return (1);
			}
		}
		while (*p1++);
		return (0);
	}

	/* regular algorithm */

	do
	{
		if (advance (p1, p2, braslist, braelist, ignore_case))
		{
			braslist [0] = p1;
			return (1);
		}
	}
	while (*p1++);

	return (0);
}

int advance(char *lp, char *ep, char *braslist[], char *braelist[],
						int ignore_case)
{
	char	*curlp;
	int	plusflg;

	plusflg = 0;		/* set to match one or more repetitions, not zero */					/* only needs setting once, since reeats drop out */
	for (;;)
		switch (*ep++)
		{
		case CCHR:
			if (ignore_case)
			{
				if (tolower(*ep++) == tolower(*lp++))
					continue;
			}
			else
			{
				if (*ep++ == *lp++)
					continue;
			}
			return (0);

		case CDOT:
			if (*lp++)
				continue;
			return (0);

		case CDOL:
			if (*lp == 0)
				continue;
			return (0);

		case CEOF:
			braelist [0] = lp;
			return (1);

		case CCL:
			if (cclass (ep, *lp++, 1, ignore_case))
			{
				ep += *ep;
				continue;
			}
			return (0);

		case NCCL:
			if (cclass (ep, *lp++, 0, ignore_case))
			{
				ep += *ep;
				continue;
			}
			return (0);

		case CBRA:
			braslist [*ep++] = lp;
			continue;

		case CKET:
			braelist [*ep++] = lp;
			continue;

		case CDOT+PLUS:
			plusflg++;
		case CDOT+STAR:
			curlp = lp;
			while (*lp++)
				;
			goto star;

		case CCHR+PLUS:
			plusflg++;
		case CCHR+STAR:
			curlp = lp;
			if (ignore_case)
			{
				while (tolower(*lp++) == tolower(*ep))
				;
			}
			else
			{
				while (*lp++ == *ep)
				;
			}
			ep++;
			goto star;

		case CCL+PLUS:
			plusflg++;
		case CCL+STAR:
			curlp = lp;
			while (cclass (ep, *lp++, 1, ignore_case))
				;
			ep += *ep;
			goto star;

		case NCCL+PLUS:
			plusflg++;
		case NCCL+STAR:
			curlp = lp;
			while (cclass (ep, *lp++, 0, ignore_case))
				;
			ep += *ep;
			goto star;

		star:
			if (plusflg)
				if (curlp == lp - 1)
					return (0);	/* zero occurrences */
				
			do
			{
				lp--;
				if (advance (lp, ep, braslist, braelist,
							ignore_case))
					return (1);
			}
			while (lp > curlp);
			return (0);

		default:
			errfun ("bad pattern");
		}
}

int cclass(char *set, int c, int af, int ignore_case)
{
	int 	n;
	char	cc;

	if (c == 0)
		return (0);
	if (ignore_case)
		c = tolower(c);
	n = *set++;
	while (--n)
		if ((cc = ignore_case ? tolower(*set++) : *set++) == RANGE)
		{
			if (*set++ <= c)
				if (c <= *set)
					return (af);
			set++;	/* skip if first part wasn't true */
			n -= 2;
		}
		else if (cc == c)
			return (af);
	return (!af);
}
